import sys
sys.path.append('C:\Program Files (x86)\Keysight\SD1\Libraries\Python')
import time
import keysightSD1
import keysight_hvi as kthvi

"""
PathWave Test Sync Executive Integration with PathWave FPGA

Copyright Keysight Technologies 2020

The goal of this application example is to show how to use HVI Port blocks placed in the sandbox of an M3xxxA module FPGA to allow communication 
between an HVI sequence designed using HVI API and custom blocks loaded into an instrument FPGA sandbox using Pathwave FPGA.

Please note that there are two types of HVI Port blocks that can be placed within the FPGA sandbox of an instrument: HVI Port Memory Maps and 
HVI Port Register Banks. Read/write of both type of HVI Port blocks is tested in this application example.

The complete list of HVI API functionality showcased in this application example is as follows:
 * Read/write data from/to an HVI sequence to/from an HVI port (Memory Map) inserted in a custom FPGA FW
 * Read/write data from/to an HVI sequence to/from an HVI port (Register bank) inserted in a custom FPGA FW
 * Read/write PXI line values from a custom FPGA FW
 * Usage of HVI Actions and Events to communicate with a custom FPGA FW

For further info please check the programming example document titled "PathWave Test Sync Executive Integration with PathWave FPGA".
"""

######################################
# Application Configuration 
######################################

class ApplicationConfig:
    """
    Defines module descriptors, configuration options and names of HVI engines, actions, triggers
    """
    def __init__(self):
        """
        Define PXI instruments options
        """
        # Define options to open the instruments. 
        # Complete option list can be found in the SD1 3.x User Guide
        self.options = 'channelNumbering=keysight'
        # Simulated HW Option
        self.hardware_simulated = True 
        """ 
        Define names of HVI engines, actions, events, triggers, registers
        """
        # HVI engine names
        self.primary_engine = "PrimaryEngine"
        self.secondary_engine = "SecondaryEngine"
        # HVI register names
        self.hvi_quit = "HVI Quit"
        self.action4_cnt = "Action4 Counter"
        self.counter_reg = "Loop Counter"
        self.mem_map = "Memory Map Value"
        self.mem_map_counter = "Memory Map Counter"
        self.pxi_values = "PXI Values"
        self.secondary_pxi_values = "Secondary PXI Values"
        self.secondary_counter_reg = "Secondary Counter"
        # HVI events and actions
        self.hvi_user_event_4 =  "FpgaUserEvent4"
        self.hvi_user_action_4 = "FpgaUserAction4"
        """
        Define names of FPGA sandbox resources
        """
        # Bitstream files generated by compiling PathWave FPGA project files
        self.primary_project_file = "../Bitfiles/HviPortExamplePrimary.k7z"
        self.secondary_project_file = "../Bitfiles/HviPortExampleSecondary.k7z"
        # Sandbox name defined by each instrument. See SD1 3.x User Guide for further info
        self.M3xxxA_sandbox = "sandbox0" # The M3xxxA_sandbox name is not arbitrary and cannot be changed 
        # FPGA Sandbox resource names
        # NOTE The FPGA resource names are not arbitrary. They correspond to the names defined in the PathWave FPGA project files
        self.num_primary_regs = 6 # number of mem. maps and registers placed in the primary PathWave FPGA project
        self.num_secondary_regs = 3 # number of mem. maps and registers placed in the secondary PathWave FPGA project
        self.memory_map = "MainEngine_Memory_1"
        self.reg_action4_cnt = "Register_Bank_HviAction4Cnt"
        self.reg_event4 = "Register_Bank_HviEvent4"
        self.reg_pxi_out = "Register_Bank_HviPxiTrigOut"
        self.reg_pxi_in = "Register_Bank_HviPxiTrigIn"
        self.secondary_reg_pxi_in = "Register_Bank_HviPxiTrigIn"
        """
        Define HW Platform
        """
        # Define module descriptors below with your instruments information
        self.module_descriptors = [
            ModuleDescriptor('M3202A', 2, 4, self.options, self.primary_engine),
            ModuleDescriptor('M3202A', 2, 15, self.options, self.secondary_engine)]
        
        # Define list of chassis containing instruments to be used by HVI
        self.chassis_list = [2]
        
        # Multi-chassis setup
        # In case of multiple chassis, chassis PXI lines need to be shared using M9031 PXI modules. 
        # See programming example documentation for more details. For example in a three-chassis setup:
        # self.M9031_descriptors = [
        #     M9031Descriptor(1, 11, 2, 10),
        #     M9031Descriptor(2, 16, 3, 9)]
        # In case of single-chassis setup please leave the list empty: M9031_descriptors = [] 
        self.M9031_descriptors = []
        
        # Assign triggers to HVI object to be used for HVI-managed synchronization, data sharing, etc
        # NOTE: In a multi-chassis setup ALL the PXI lines listed below need to be shared among each M9031 board pair by means of SMB cable connections
        self.pxi_sync_trigger_resources = [
            kthvi.TriggerResourceId.PXI_TRIGGER0,
            kthvi.TriggerResourceId.PXI_TRIGGER1,
            kthvi.TriggerResourceId.PXI_TRIGGER2,
            kthvi.TriggerResourceId.PXI_TRIGGER3]

######################################
# Main Program
######################################

def main():
    ###################################################
    # Open PXI modules, load PathWave FPGA bitfiles
    ###################################################
    # Load configuration
    config = ApplicationConfig()

    # Checks flags defining if the code executes on HW/Simulation, SD1
    if config.hardware_simulated:
        print("Code running in Simulation Mode")

    # Open instruments
    module_dict = open_modules(config)

    # Load bitfiles generated with PathWave FPGA
    if not config.hardware_simulated:
        load_fpga_bitfiles(module_dict, config)

    ######################################
    # System definition
    ######################################
    my_system = kthvi.SystemDefinition("MySystem") 
    # Define your system, HW platform and HVI resources
    define_hvi_resources(my_system, module_dict, config)
    
    ######################################
    # Program HVI sequences
    ######################################
    print("Programming the HVI sequences...")
    # Create sequencer object
    sequencer = kthvi.Sequencer("MySequencer", my_system)
    # Program HVI sequences using the sequencer object
    program_fpga_interaction_sequence(sequencer, config)
    # Export the programmed sequence to text
    export_hvi_seqeunces(sequencer)

    ########################################
    # Compile, Load to HW, Run HVI Sequence
    ########################################
    # Compile HVI sequences
    hvi = sequencer.compile()
    print(hvi.compile_status.to_string())
    print("HVI Compiled")
    print("This HVI application needs to reserve {} PXI trigger resources to execute".format(len(hvi.compile_status.sync_resources)))
    
    # Load HVI to HW: load sequences, configure actions/triggers/events, lock resources, etc.
    hvi.load_to_hw()    
    print("HVI Loaded to HW")

    # Execute HVI in non-blocking mode 
    # This mode allows SW execution to interact with HVI execution
    hvi.run(hvi.no_wait)
    print("HVI Running...")

    # User-controlled HVI execution
    if config.hardware_simulated:
        print("Simulation completed successfully")
    else:
        interact_with_hvi(hvi, config)

    # Release HW resources once HVI execution is completed
    hvi.release_hw()
    print("Releasing HW...")

    # Close all modules at the end of the execution
    for engine_name in module_dict:
        module_dict[engine_name].instrument.close()
    print("Modules closed\n")

########################################################
# 2nd Level: Functions to Define, Program, Execute HVI
#######################################################

def load_fpga_bitfiles(module_dict, config):
    """
    Loads the bitstream to the sandbox of each PXI modules and displays a summary of the FPGA sandbox resources
    """
    # Primary module, secondary module
    primary_module = module_dict[config.primary_engine].instrument
    secondary_module = module_dict[config.secondary_engine].instrument
    
    # Load custom FPGA FW designed using Keysight Pathwave FPGA
    error = primary_module.FPGAload(config.primary_project_file)
    if error < 0:
        raise Exception("primary_module.FPGAload() error! Error code: {} Error message: {}".format(error, keysightSD1.SD_Error.getErrorMessage(error)))
    error = secondary_module.FPGAload(config.secondary_project_file)
    if error < 0:
        raise Exception("secondary_module.FPGAload() error! Error code: {} Error message: {}".format(error, keysightSD1.SD_Error.getErrorMessage(error)))

    #Get primary module's registers info
    print("Registers contained in Primary Module:")
    registers = primary_module.FPGAgetSandBoxRegisters(config.num_primary_regs)

    #Print the register properties in register list
    for register in registers:
        print("Register name: {}".format(register.Name))
        print("Register size [bytes]: {}".format(register.Length))
        print("Register address offset [bytes]: {}".format(register.Address))
        print("Register access: {}".format(register.AccessType))
    print()

    #Get secondary module's registers info
    print("Registers contained in Secondary Module:")
    registers = secondary_module.FPGAgetSandBoxRegisters(config.num_secondary_regs)

    #Print the register properties in register list
    for register in registers:
        print("Register name: {}".format(register.Name))
        print("Register size [bytes]: {}".format(register.Length))
        print("Register address offset [bytes]: {}".format(register.Address))
        print("Register access: {}".format(register.AccessType))
    print()


def define_hvi_resources(sys_def, module_dict, config):
    """
    Configures all the necessary resources for the HVI application to execute: HW platform, engines, actions, triggers, etc. 
    """
    # Define HW platform: chassis, interconnections, PXI trigger resources, synchronization, HVI clocks
    define_hw_platform(sys_def, config)
    
    # Define all the HVI engines to be included in the HVI
    define_hvi_engines(sys_def, module_dict)
    
    # Define FPGA actions, events and other configurations
    define_fpga_resources(sys_def, module_dict, config)


def define_hw_platform(sys_def, config):
    """
     Define HW platform: chassis, interconnections, PXI trigger resources, synchronization, HVI clocks 
    """
    # Add chassis resources 
    # For multi-chassis setup details see programming example documentation
    for chassis_number in config.chassis_list:
        if config.hardware_simulated:
            sys_def.chassis.add_with_options(chassis_number, 'Simulate=True,DriverSetup=model=M9018B,NoDriver=True')
        else:
            sys_def.chassis.add(chassis_number)

    # Add M9031 modules for multi-chassis setups
    if config.M9031_descriptors:
        interconnects = sys_def.interconnects
        for descriptor in config.M9031_descriptors:
            interconnects.add_M9031_modules(descriptor.chassis_1, descriptor.slot_1, descriptor.chassis_2, descriptor.slot_2)

    # Assign the defined PXI trigger resources
    sys_def.sync_resources = config.pxi_sync_trigger_resources
    
    # Assign clock frequencies that are outside the set of the clock frequencies of each HVI engine
    # Use the code line below if you want the application to be in sync with the 10 MHz clock
    sys_def.non_hvi_core_clocks = [10e6]


def define_hvi_engines(sys_def, module_dict):
    """
    Define all the HVI engines to be included in the HVI
    """
    # For each instrument to be used in the HVI application add its HVI Engine to the HVI Engine Collection 
    for engine_name, module in zip(module_dict.keys(), module_dict.values()):
        sys_def.engines.add(module.instrument.hvi.engines.main_engine, engine_name)


def define_fpga_resources(sys_def, module_dict, config):
    """ 
    Define FPGA actions, events and other configurations 
    """
    # Primary module, secondary module
    primary_module = module_dict[config.primary_engine].instrument
    secondary_module = module_dict[config.secondary_engine].instrument

    # Events: add FpgaUserEvent4 to the list of events of the primary engine
    fpga_user_event4 = primary_module.hvi.events.fpga_user_4
    sys_def.engines[config.primary_engine].events.add(fpga_user_event4, config.hvi_user_event_4)

    # Actions: add FpgaUserAction4 to the list of actions of the primary engine
    fpga_user_action4 = primary_module.hvi.actions.fpga_user_4
    sys_def.engines[config.primary_engine].actions.add(fpga_user_action4, config.hvi_user_action_4)

    # Get engine sandbox
    sandbox_name = config.M3xxxA_sandbox
    primary_sandbox = sys_def.engines[config.primary_engine].fpga_sandboxes[sandbox_name]
    secondary_sandbox = sys_def.engines[config.secondary_engine].fpga_sandboxes[sandbox_name]
    # Load to the sandboxes .k7z project created using Pathwave FPGA
    # This operation is necessary for HVI to list all the FPGA blocks contrined in the designed FPGA FW
    primary_sandbox.load_from_k7z(config.primary_project_file)
    secondary_sandbox.load_from_k7z(config.secondary_project_file)
    
    # Enable PXI lines to be written from the FPGA sandbox of primary engine only using FPGATriggerOutConfig()
    # NOTE: Only one PXI module per segment shall be allowed to write backplane PXI lines. It would cause conflicts and misbehavior to configure the PXI lines for the secondary engine also 
    primary_module.FPGATriggerConfig(externalSource=keysightSD1.SD_TriggerExternalSources.TRIGGER_PXI4, direction =keysightSD1.SD_FpgaTriggerDirection.INOUT, polarity= keysightSD1.SD_TriggerPolarity.ACTIVE_LOW, syncMode = keysightSD1.SD_SyncModes.SYNC_NONE, delay5Tclk=0)
    primary_module.FPGATriggerConfig(externalSource=keysightSD1.SD_TriggerExternalSources.TRIGGER_PXI5, direction =keysightSD1.SD_FpgaTriggerDirection.INOUT, polarity= keysightSD1.SD_TriggerPolarity.ACTIVE_LOW, syncMode = keysightSD1.SD_SyncModes.SYNC_NONE, delay5Tclk=0)
    primary_module.FPGATriggerConfig(externalSource=keysightSD1.SD_TriggerExternalSources.TRIGGER_PXI6, direction =keysightSD1.SD_FpgaTriggerDirection.INOUT, polarity= keysightSD1.SD_TriggerPolarity.ACTIVE_LOW, syncMode = keysightSD1.SD_SyncModes.SYNC_NONE, delay5Tclk=0)
    primary_module.FPGATriggerConfig(externalSource=keysightSD1.SD_TriggerExternalSources.TRIGGER_PXI7, direction =keysightSD1.SD_FpgaTriggerDirection.INOUT, polarity= keysightSD1.SD_TriggerPolarity.ACTIVE_LOW, syncMode = keysightSD1.SD_SyncModes.SYNC_NONE, delay5Tclk=0)

    secondary_module.FPGATriggerConfig(externalSource=keysightSD1.SD_TriggerExternalSources.TRIGGER_PXI4, direction =keysightSD1.SD_FpgaTriggerDirection.IN, polarity= keysightSD1.SD_TriggerPolarity.ACTIVE_LOW, syncMode = keysightSD1.SD_SyncModes.SYNC_NONE, delay5Tclk=0)
    secondary_module.FPGATriggerConfig(externalSource=keysightSD1.SD_TriggerExternalSources.TRIGGER_PXI5, direction =keysightSD1.SD_FpgaTriggerDirection.IN, polarity= keysightSD1.SD_TriggerPolarity.ACTIVE_LOW, syncMode = keysightSD1.SD_SyncModes.SYNC_NONE, delay5Tclk=0)
    secondary_module.FPGATriggerConfig(externalSource=keysightSD1.SD_TriggerExternalSources.TRIGGER_PXI6, direction =keysightSD1.SD_FpgaTriggerDirection.IN, polarity= keysightSD1.SD_TriggerPolarity.ACTIVE_LOW, syncMode = keysightSD1.SD_SyncModes.SYNC_NONE, delay5Tclk=0)
    secondary_module.FPGATriggerConfig(externalSource=keysightSD1.SD_TriggerExternalSources.TRIGGER_PXI7, direction =keysightSD1.SD_FpgaTriggerDirection.IN, polarity= keysightSD1.SD_TriggerPolarity.ACTIVE_LOW, syncMode = keysightSD1.SD_SyncModes.SYNC_NONE, delay5Tclk=0)


def program_fpga_interaction_sequence(sequencer, config):
    """
    This method programs the HVI sequence of this application.
    Different HVI statements are encapsulated as much as possible in separated SW methods to help users visualize
    the programmed HVI sequences.
    The programming example documentation on www.keysight.com contains an HVI diagram that graphically represents the programmed HVI sequence.
    """
    # Define registers within the scope of the outmost sync sequence
    define_registers(sequencer, config)

    # Define Sync While condition
    hvi_quit = sequencer.sync_sequence.scopes[config.primary_engine].registers[config.hvi_quit]
    sync_while_condition = kthvi.Condition.register_comparison(hvi_quit, kthvi.ComparisonOperator.NOT_EQUAL_TO, 1)
    # Add Sync While statement
    sync_while = sequencer.sync_sequence.add_sync_while("User-controlled sync loop", 90, sync_while_condition)
    # Program sync loop 
    program_sync_loop(sync_while.sync_sequence, config)


def define_registers(sequencer, config):
    """
    Defines all registers for each HVI engine in the scope af the global sync sequence 
    """
    # Define registers for primary engine
    primary_engine_register_collection = sequencer.sync_sequence.scopes[config.primary_engine].registers
    hvi_quit = primary_engine_register_collection.add(config.hvi_quit, kthvi.RegisterSize.SHORT)
    hvi_quit.initial_value = 0
    action4_cnt = primary_engine_register_collection.add(config.action4_cnt, kthvi.RegisterSize.SHORT)
    action4_cnt.initial_value = 0
    counter_reg = primary_engine_register_collection.add(config.counter_reg, kthvi.RegisterSize.SHORT)
    counter_reg.initial_value = 0
    mem_map = primary_engine_register_collection.add(config.mem_map, kthvi.RegisterSize.SHORT)
    mem_map.initial_value = 0
    mem_map_counter = primary_engine_register_collection.add(config.mem_map_counter, kthvi.RegisterSize.SHORT)
    mem_map_counter.initial_value = 1000
    pxi_values = primary_engine_register_collection.add(config.pxi_values, kthvi.RegisterSize.SHORT)
    pxi_values.initial_value = 0
    
    # Define registers for primary engine
    secondary_engine_register_collection = sequencer.sync_sequence.scopes[config.secondary_engine].registers
    secondary_counter_reg = secondary_engine_register_collection.add(config.secondary_counter_reg, kthvi.RegisterSize.SHORT)
    secondary_counter_reg.initial_value = 0
    secondary_pxi_values = secondary_engine_register_collection.add(config.secondary_pxi_values, kthvi.RegisterSize.SHORT)
    secondary_pxi_values.initial_value = 0


def program_sync_loop(sync_sequence, config):
    """
    Adds and programs the outmost Sync While statement of the HVI Sync Sequence
    """
    # Add 1st Sync Multi-Sequence Block to the Sync While sequence
    sync_block_1 = sync_sequence.add_sync_multi_sequence_block("FPGA Read/Write Operations", 260)
    # Add and program 1st Sync Multi-Sequence Block
    program_fpga_operations(sync_block_1, config)

    # In this Sync Multi-Seq. Block both primary and secondary modules read into an HVI register the PXI lines containing the action4 counter value.
    # NOTE: It is critical for the reading operation to be successful to allow enough delay from the previous SMSB where the action4 counter
    # value was written to PXI lines. This is done using the pxi_propagation_delay Python variable. 
    # The necessary pxi_propagation_delay value depends on the module position (segments, chassis).
    # When both primary and secondary modules are placed in the same segment, or adjacent segments, pxi_propagation_delay = 100 ns shall be enough.
    # The more the secondary and primary module are far away, the more the pxi_propagation_delay value shall be increased.
    pxi_propagation_delay = 300 #[ns]
    # Add 2nd Sync Multi-Sequence Block (SMSB) to the Sync While sequence
    sync_block_2 = sync_sequence.add_sync_multi_sequence_block("Wait for HVI_UserEvent4 and Execute HVI_UserAction4", pxi_propagation_delay)
    # Add and program 2nd Sync Multi-Sequence Block
    program_wait_event_block(sync_block_2, config)

    # Add a 3rd Sync Multi-Sequence Block to the Sync While sequence
    sync_block_3 = sync_sequence.add_sync_multi_sequence_block("Increment counter registers", 10)
    # Add and program 3rd Sync Multi-Sequence Block
    program_increment_counters(sync_block_3, config)


def program_fpga_operations(sync_block, config):
    """
    Programs a Sync Multi-Sequence Block (SMSB) statement to execute FPGA Read/Write Operations
    """
    # Local sequence objects within the SMSB
    primary_sequence = sync_block.sequences[config.primary_engine]

    #Previously defined registers
    action4_cnt = primary_sequence.scope.registers[config.action4_cnt]
    mem_map_counter = primary_sequence.scope.registers[config.mem_map_counter]
    mem_map = primary_sequence.scope.registers[config.mem_map]
    #Previously defined FPGA resources
    fpga_reg_action4_cnt = primary_sequence.engine.fpga_sandboxes[config.M3xxxA_sandbox].fpga_registers[config.reg_action4_cnt]
    fpga_reg_pxi_out = primary_sequence.engine.fpga_sandboxes[config.M3xxxA_sandbox].fpga_registers[config.reg_pxi_out]
    fpga_memory_map = primary_sequence.engine.fpga_sandboxes[config.M3xxxA_sandbox].fpga_memory_maps[config.memory_map]
    
    # Read FPGA Register Register_Bank_HviAction4Cnt
    readFpgaReg0 = primary_sequence.add_instruction("Read FPGA Register_Bank_HviAction4Cnt", 10, primary_sequence.instruction_set.fpga_register_read.id)
    readFpgaReg0.set_parameter(primary_sequence.instruction_set.fpga_register_read.destination.id, action4_cnt)
    readFpgaReg0.set_parameter(primary_sequence.instruction_set.fpga_register_read.fpga_register.id, fpga_reg_action4_cnt)

    # Write FPGA Register Register_Bank_HviPxiTrigOut
    # Register_Bank_HviPxiTrigOut is connected to PXI lines Outputs. The value written to the FPGA register will be written to PXI lines
    # NOTE: Please allow at least 60 ns between these instructions to ensure the HVI register action4_cnt is updated before writing its content to PXI lines
    writeFpgaReg0 = primary_sequence.add_instruction("Write FPGA Register_Bank_HviPxiTrigOut", 60, primary_sequence.instruction_set.fpga_register_write.id)
    writeFpgaReg0.set_parameter(primary_sequence.instruction_set.fpga_register_write.fpga_register.id, fpga_reg_pxi_out)
    writeFpgaReg0.set_parameter(primary_sequence.instruction_set.fpga_register_write.value.id, action4_cnt)

    # Write Memory Map
    # At each iteration a different value is written to the memory map
    writeMemoryMap = primary_sequence.add_instruction("Write FPGA Memory Map", 10, primary_sequence.instruction_set.fpga_array_write.id)
    writeMemoryMap.set_parameter(primary_sequence.instruction_set.fpga_array_write.fpga_memory_map.id, fpga_memory_map)
    writeMemoryMap.set_parameter(primary_sequence.instruction_set.fpga_array_write.value.id, mem_map_counter)
    writeMemoryMap.set_parameter(primary_sequence.instruction_set.fpga_array_write.fpga_memory_map_offset.id, 0)

    # Read Memory Map
    # Reads the value that was written to the block RAM connected to the memory map
    # NOTE: Please allow at least 30 ns between these instructions to ensure data is written correctly through the memory map before you read it back
    readMemoryMap = primary_sequence.add_instruction("Read FPGA Memory Map", 30, primary_sequence.instruction_set.fpga_array_read.id)
    readMemoryMap.set_parameter(primary_sequence.instruction_set.fpga_array_read.fpga_memory_map.id, fpga_memory_map)
    readMemoryMap.set_parameter(primary_sequence.instruction_set.fpga_array_read.destination.id, mem_map)
    readMemoryMap.set_parameter(primary_sequence.instruction_set.fpga_array_read.fpga_memory_map_offset.id, 0)

    return


def program_wait_event_block(sync_block, config):
    """
    Programs a Sync Multi-Sequence Block (SMSB) statement to wait for a sandbox event and execute a sandbox action
    """
    # Local sequence objects within the SMSB
    primary_sequence = sync_block.sequences[config.primary_engine]
    secondary_sequence = sync_block.sequences[config.secondary_engine]

    # Previously defined registers
    pxi_values = primary_sequence.scope.registers[config.pxi_values]
    secondary_pxi_values = secondary_sequence.scope.registers[config.secondary_pxi_values]
    # Previously defined FPGA resources
    fpga_reg_pxi_in = primary_sequence.engine.fpga_sandboxes[config.M3xxxA_sandbox].fpga_registers[config.reg_pxi_in]
    fpga_secondary_reg_pxi_in = secondary_sequence.engine.fpga_sandboxes[config.M3xxxA_sandbox].fpga_registers[config.secondary_reg_pxi_in]

    # Both Primary and Secondary modules read PXI liness logical values from FPGA Register Register_Bank_HviPxiTrigIn
    # In the sandbox of both primary and secondary modules sandbox FPGA Register Register_Bank_HviPxiTrigIn is connected to PXI lines inputs
    readFpgaReg0 = primary_sequence.add_instruction("Read Register_Bank_HviPxiTrigIn", 10, primary_sequence.instruction_set.fpga_register_read.id)
    readFpgaReg0.set_parameter(primary_sequence.instruction_set.fpga_register_read.destination.id, pxi_values)
    readFpgaReg0.set_parameter(primary_sequence.instruction_set.fpga_register_read.fpga_register.id, fpga_reg_pxi_in)
    # Read PXI lines from FPGA Register Register_Bank_HviPxiTrigIn in secondary instrument 
    readSecondaryFpgaReg2 = secondary_sequence.add_instruction("Read Register_Bank_HviPxiTrigIn", 10, secondary_sequence.instruction_set.fpga_register_read.id)
    readSecondaryFpgaReg2.set_parameter(secondary_sequence.instruction_set.fpga_register_read.destination.id, secondary_pxi_values)
    readSecondaryFpgaReg2.set_parameter(secondary_sequence.instruction_set.fpga_register_read.fpga_register.id, fpga_secondary_reg_pxi_in)

    # Wait for FPGA_User_Event4
    # Define the condition for the wait statement
    wait_condition = kthvi.Condition.event(primary_sequence.engine.events[config.hvi_user_event_4])
    # Add wait statement
    waitEvent = primary_sequence.add_wait("Wait for FPGA_User_Event4", 10, wait_condition)
    waitEvent.set_mode(kthvi.WaitMode.TRANSITION, kthvi.SyncMode.IMMEDIATE)

    # Action execute instruction: execute action 4
    instAction4 = primary_sequence.add_instruction("Execute Action 4", 20, primary_sequence.instruction_set.action_execute.id)
    instAction4.set_parameter(primary_sequence.instruction_set.action_execute.action.id, primary_sequence.engine.actions[config.hvi_user_action_4])

    return


def program_increment_counters(sync_block, config):
    """
    Programs a Sync Multi-Sequence Block (SMSB) statement to increment register counters
    """
    # Local sequence objects within the SMSB
    primary_sequence = sync_block.sequences[config.primary_engine]
    secondary_sequence = sync_block.sequences[config.secondary_engine]

    #Previously defined registers
    counter_reg = primary_sequence.scope.registers[config.counter_reg]
    mem_map_counter = primary_sequence.scope.registers[config.mem_map_counter]
    secondary_counter_reg = secondary_sequence.scope.registers[config.secondary_counter_reg]

    # Increment memory map counter
    instr = primary_sequence.add_instruction("Increment memory map counter", 10, primary_sequence.instruction_set.add.id)
    instr.set_parameter(primary_sequence.instruction_set.add.left_operand.id, mem_map_counter)
    instr.set_parameter(primary_sequence.instruction_set.add.right_operand.id, 1)
    instr.set_parameter(primary_sequence.instruction_set.add.destination.id, mem_map_counter)

    # Increment counter register
    instr = primary_sequence.add_instruction("Increment counter register", 10, primary_sequence.instruction_set.add.id)
    instr.set_parameter(primary_sequence.instruction_set.add.left_operand.id, counter_reg)
    instr.set_parameter(primary_sequence.instruction_set.add.right_operand.id, 1)
    instr.set_parameter(primary_sequence.instruction_set.add.destination.id, counter_reg)

    # Increment secondary counter register
    instr = secondary_sequence.add_instruction("Increment Secondary counter register", 10, secondary_sequence.instruction_set.add.id)
    instr.set_parameter(secondary_sequence.instruction_set.add.left_operand.id, secondary_counter_reg)
    instr.set_parameter(secondary_sequence.instruction_set.add.right_operand.id, 1)
    instr.set_parameter(secondary_sequence.instruction_set.add.destination.id, secondary_counter_reg)


def export_hvi_seqeunces(sequencer):
    """
    Exports the programmed HVI sequences to text format
    """
    # Generate HVI sequence description text
    output = sequencer.sync_sequence.to_string(kthvi.OutputFormat.DEBUG)
    print("Programmed HVI sequences exported to file")
    
    # Write file
    text_file = open("../TextOutput/HviPwFpgaIntegration.txt", "w")
    text_file.write(output)
    text_file.close()


def interact_with_hvi(hvi, config):
    """
    This function allows users to interact with the HVI execution from SW execution
    """ 
    # Primary engine registers
    primary_engine_scope = hvi.sync_sequence.scopes[config.primary_engine]
    action4_cnt = primary_engine_scope.registers[config.action4_cnt]
    mem_map = primary_engine_scope.registers[config.mem_map]
    mem_map_counter_reg = primary_engine_scope.registers[config.mem_map_counter]
    pxi_values = primary_engine_scope.registers[config.pxi_values]
    counter_reg = primary_engine_scope.registers[config.counter_reg]
    hvi_quit = primary_engine_scope.registers[config.hvi_quit]
    # Secondary engine registers
    secondary_engine_scope = hvi.sync_sequence.scopes[config.secondary_engine]
    secondary_pxi_values = secondary_engine_scope.registers[config.secondary_pxi_values]
    secondary_counter_reg = secondary_engine_scope.registers[config.secondary_counter_reg]

    # User controlled loop
    primary_sandbox = hvi.sync_sequence.engines[config.primary_engine].fpga_sandboxes[config.M3xxxA_sandbox]
    while True:
        print("N. of User Actions counted at primary module: action4_cnt = {}".format(action4_cnt.read()))
        print("Value written to the FPGA Memory Map: mem_map = {}".format(mem_map.read()))
        print("Value read by Primary Module from FPGA PXI inputs: pxi_values = {}".format(pxi_values.read()))
        print("Value read by Secondary Module from FPGA PXI inputs: secondary_pxi_values = {}".format(secondary_pxi_values.read()))
        print("Counter value: counter_reg = {}".format(counter_reg.read()))
        print("Secondary counter value: secondary_counter_reg = {}".format(secondary_counter_reg.read()))
        print("Mem. Map counter value: mem_map_counter_reg = {}".format(mem_map_counter_reg.read()))
        print("Press enter to trigger a User Event and execute a User Action, press q to exit")
        key = input()

        # Generate an event or quit
        if key == 'q':
            hvi_quit.write(1)
            time.sleep(.001)
            triggerFpgaUserEvent(primary_sandbox, config)
            break
        else:
            triggerFpgaUserEvent(primary_sandbox, config)


def triggerFpgaUserEvent(sandbox, config):
    """
    Triggers an FPGA_Event4 from the sandbox by writing an HVI sandbox register connected to Event4
    """
    # Write a 1-0 to register "Register_Bank_HviEvent4" within an HVI Port Register Bank
    # Please update the registerName to the actual name of your register if necessary 
    regEvent = sandbox.fpga_registers[config.reg_event4]
    regEvent.write(1)
    regEvent.write(0)

#################################################################
# 3rd Level: Classes and Functions to Use SD1/M3xxxA Instruments
#################################################################

class ModuleDescriptor:
    "Descriptor for module objects"
    def __init__(self, model_number, chassis_number, slot_number, options, engine_name):
        self.model_number = model_number
        self.chassis_number = chassis_number
        self.slot_number = slot_number
        self.options = options
        self.engine_name = engine_name


class M9031Descriptor:
    "Describes the interconnection between each pair of M9031A modules"
    def __init__(self, first_M9031_chassis_number, first_M9031_slot_number, second_M9031_chassis_number, second_M9031_slot_number):
        self.chassis_1 = first_M9031_chassis_number
        self.slot_1 = first_M9031_slot_number
        self.chassis_2 = second_M9031_chassis_number
        self.slot_2 = second_M9031_slot_number


class Module:
    "Class defining a modular instrument object and its properties"
    def __init__(self, instrument_object, num_channels):
        self.instrument = instrument_object
        self.num_channels = num_channels


def open_modules(config):
    """ 
    Opens and creates all the necessary instrument objects.
    Returns a dictionary of module objects whose keys are the HVI engine names.
    Please check SD1 3.x User Manual for options to open the instrument objects
    """
    # Initialize output variables
    num_modules = 0
    module_dict = {} # dictionary of modules

    # Open SD1 instrument objects
    for descriptor in config.module_descriptors:
        if descriptor.model_number == "M3102A":
            instr_obj = keysightSD1.SD_AIN()
        else:
            instr_obj = keysightSD1.SD_AOU()
        instr_obj_options = descriptor.options + ',simulate=true' if config.hardware_simulated else descriptor.options
        id = instr_obj.openWithOptions(descriptor.model_number, descriptor.chassis_number, descriptor.slot_number, instr_obj_options)
        if id < 0:
            raise Exception("Error opening instrument in chassis: {}, slot: {}! Error code: {} - {}. Exiting...".format(descriptor.chassis_number, descriptor.slot_number, id, keysightSD1.SD_Error.getErrorMessage(id)))
        nCh = instr_obj.getOptions("channels")
        if nCh == "CH2":
            num_channels = 2
        elif nCh == "CH4":
            num_channels = 4
        else:
            raise Exception("PXI module in chassis {}, slot {} returned number of channels = {} which is incorrect. Exiting... ".format(instr_obj.getChassis(), instr_obj.getSlot(), nCh))
        module_dict[descriptor.engine_name] = Module(instr_obj, num_channels)
        num_modules += 1
    
    return module_dict


if __name__ == '__main__':
    main()
